searchNodes=[{"doc":"Drop-in replacement of :elixir_make","ref":"Mix.Tasks.Compile.FennecPrecompile.html","title":"Mix.Tasks.Compile.FennecPrecompile","type":"task"},{"doc":"Callback implementation for Mix.Task.run/1 .","ref":"Mix.Tasks.Compile.FennecPrecompile.html#run/1","title":"Mix.Tasks.Compile.FennecPrecompile.run/1","type":"function"},{"doc":"A task responsible for downloading the precompiled NIFs for a given module. This task must only be used by package creators who want to ship the precompiled NIFs. The goal is to download the precompiled packages and generate a checksum to check-in alongside the project in the the Hex repository. This is done by passing the --all flag. You can also use the --only-local flag to download only the precompiled package for use during development. You can use the --ignore-unavailable flag to ignore any NIFs that are not available. This is useful when you are developing a new NIF that does not support all platforms. This task also accept the --print flag to print the checksums.","ref":"Mix.Tasks.Fennec.Fetch.html","title":"Mix.Tasks.Fennec.Fetch","type":"task"},{"doc":"Download and use precompiled NIFs safely with checksums. Fennec Precompile is a tool for library maintainers that use :elixir_make and wish to ship precompiled binaries. This tool aims to be a drop-in replacement for :elixir_make . It helps by removing the need to have the C/C++ compiler and other dependencies installed in the user's machine. Check the Precompilation Guide for details. Options to set in the project function of the mix.exs file. :fennec_base_url . Required. Specifies the base download URL of the precompiled binaries. :fennec_nif_filename . Optional. Specifies the name of the precompiled binary file, excluding the file extension. :fennec_force_build . Optional. Indicates whether to force the app to be built. The value of this option will always be true for pre-releases (like &quot;2.1.0-dev&quot;). When this value is false and there are no local or remote precompiled binaries, a compilation error will be raised. :fennec_force_build_args . Optional. Defaults to [] . This option will be used when :force_build is true . The optional compiliation args will be forwarded to :elixir_make . :fennec_force_build_using_zig . Optional. Defaults to false . This option will be used when :force_build is true . Set this option to true to always using zig as the C/C++ compiler. :fennec_targets . Optional. A list of targets supported by Zig for which precompiled assets are avilable. By default the following targets are configured: on macOS x86_64-macos x86_64-linux-gnu x86_64-linux-musl x86_64-windows-gnu aarch64-macos aarch64-linux-gnu aarch64-linux-musl riscv64-linux-musl on Linux x86_64-linux-gnu x86_64-linux-musl x86_64-windows-gnu aarch64-linux-gnu aarch64-linux-musl riscv64-linux-musl :fennec_targets in the project will only be used in the following cases: When :fennec_force_build is set to true . In this case, the :targets acts as a list of compatible targets in terms of the source code. For example, NIFs that are specifically written for ARM64 Linux will fail to compile for other OS or CPU architeture. If the source code is not compatible with the current node, the build will fail. When :fennec_force_build is set to false . In this case, the :targets acts as a list of available targets of the precompiled binaries. If there is no match with the current node, no precompiled NIF will be downloaded and the app will fail to start.","ref":"Mix.Tasks.Fennec.Precompile.html","title":"Mix.Tasks.Fennec.Precompile","type":"task"},{"doc":"","ref":"Mix.Tasks.Fennec.Precompile.html#build_native/1","title":"Mix.Tasks.Fennec.Precompile.build_native/1","type":"function"},{"doc":"","ref":"Mix.Tasks.Fennec.Precompile.html#build_native_using_zig/1","title":"Mix.Tasks.Fennec.Precompile.build_native_using_zig/1","type":"function"},{"doc":"","ref":"Mix.Tasks.Fennec.Precompile.html#build_with_targets/3","title":"Mix.Tasks.Fennec.Precompile.build_with_targets/3","type":"function"},{"doc":":fennec_precompile aims to be a drop-in replacement for :elixir_make . It provides a convenient way to precompile your NIF app for all the supported platforms. This is mostly useful for the following reasons: when a working C/C++ toolchain is not available (e.g, running livebook on some embedded devices, or running in nerves environment, where a C/C++ compiler is not included) a working C/C++ compiler won't be a strict requirement for using your NIF app. save time on compiling. In the following sections, I will walk you through how to use :fennec_precompile to precompile an example app, :fennec_example .","ref":"precompilation_guide.html","title":"Precompilation guide","type":"extras"},{"doc":"We start by creating a new app, say fennec_precompile . $ mix new fennec_example * creating README.md * creating .formatter.exs * creating .gitignore * creating mix.exs * creating lib * creating lib/fennec_example.ex * creating test * creating test/test_helper.exs * creating test/fennec_example_test.exs Your Mix project was created successfully. You can use &quot;mix&quot; to compile it, test it, and more: cd fennec_example mix test Run &quot;mix help&quot; for more commands.","ref":"precompilation_guide.html#create-a-new-app","title":"Precompilation guide - Create a new app","type":"extras"},{"doc":"In the mix.exs file, we add :fennec_precompile to deps . Also, note that :fennec_precompile should not be added to the :compilers list. defmodule FennecExample do use Mix.Project @version &quot;0.1.0&quot; def project do [ # ... compilers : [ :fennec_precompile ] ++ Mix . compilers ( ) fennec_base_url : &quot;https://github.com/cocoa-xu/fennec_example/downloads/releases/v \#{ @version } &quot; , fennec_nif_filename : &quot;nif&quot; # ... ] end def deps do [ # ... { :fennec_precompile , &quot;~&gt; 0.1&quot; } # ... ] end end","ref":"precompilation_guide.html#add-fennec_precompile-to-the-mix-exs","title":"Precompilation guide - Add :fennec_precompile to the mix.exs","type":"extras"},{"doc":"Then we write a toy NIF function that returns 'hello world' . $ mkdir -p c_src $ cat &lt;&lt;EOF | tee c_src/fennec_example.c #include &lt;stdio.h&gt; #include &lt;erl_nif.h&gt; static ERL_NIF_TERM hello_world(ErlNifEnv* env, int argc, const ERL_NIF_TERM argv[]) { return enif_make_string(env, &quot;hello world&quot;, ERL_NIF_LATIN1); } static ErlNifFunc nif_funcs[] = { {&quot;hello_world&quot;, 0, hello_world, 0} }; ERL_NIF_INIT(fennec_example, nif_funcs, NULL, NULL, NULL, NULL); EOF # note that here the output file is named as `nif.so` $ cat &lt;&lt;EOF | tee Makefile PRIV_DIR = \\$(MIX_APP_PATH)/priv NIF_SO = \\$(PRIV_DIR)/nif.so C_SRC = \\$(shell pwd)/c_src CFLAGS += -shared -std=c11 -O3 -fPIC -I\\$(ERTS_INCLUDE_DIR) UNAME_S := \\$(shell uname -s) ifeq (\\$(UNAME_S),Darwin) CFLAGS += -undefined dynamic_lookup -flat_namespace -undefined suppress endif .DEFAULT_GLOBAL := build build: \\$(NIF_SO) \\$(NIF_SO): \\$(C_SRC)/fennec_example.c @ mkdir -p \\$(PRIV_DIR) \\$(CC) \\$(CFLAGS) \\$(C_SRC)/fennec_example.c -o \\$(NIF_SO) EOF # `nif_filename` is set to `nif`, which is the name of the NIF file excluding the extension. $ cat &lt;&lt;EOF | tee lib/fennec_example.ex defmodule :fennec_example do @moduledoc false @on_load :load_nif def load_nif do nif_file = &#39;\#{:code.priv_dir(:fennec_example)}/nif&#39; case :erlang.load_nif(nif_file, 0) do :ok -&gt; :ok {:error, {:reload, _}} -&gt; :ok {:error, reason} -&gt; IO.puts(&quot;Failed to load nif: \#{reason}&quot;) end end def hello_world(), do: :erlang.nif_error(:not_loaded) end EOF $ cat &lt;&lt;EOF | tee test/fennec_example_test.ex defmodule FennecExampleTest do use ExUnit.Case test &quot;greets the world&quot; do assert :fennec_example.hello_world() == &#39;hello world&#39; end end EOF In the mix.exs file, we passed two options: :fennec_base_url . Required. Specifies the base download URL of the precompiled binaries. :fennec_nif_filename . Required. Specifies the name of the precompiled binary file, excluding the file extension. For all available options, please refer to Mix.Tasks.Fennec.Precompile . All of these values can be overridden by the user in the config/config.exs file. For instance, To avoid supply chain attack or to speed up the deployment, the user can redirect to their trusted server. import Config config :fennec_precompile , :config , fennec_example : [ fennec_base_url : &quot;https://cdn.example.com/fennec_example&quot; , fennec_force_build : false ]","ref":"precompilation_guide.html#write-nif-code","title":"Precompilation guide - Write NIF code","type":"extras"},{"doc":"To test the NIF code locally, first we need to compile for the host platform. # this is equivalent to `mix compile.elixir_make` $ mix compile.fennec_precompile ==&gt; fennec_example cc -shared -std=c11 -O3 -fPIC -I/usr/local/lib/erlang/erts-13.0/include -undefined dynamic_lookup -flat_namespace -undefined suppress /Users/cocoa/Git/fennec_example/c_src/fennec_example.c -o /Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so Of course, you can also use zig as the C/C++ compiler&lt;sup&gt; 1 &lt;/sup&gt;. # set environment variable `FENNEC_PRECOMPILE_ALWAYS_USE_ZIG` to `true` $ export FENNEC_PRECOMPILE_ALWAYS_USE_ZIG=true $ mix compile 20:42:07.566 [debug] Current compiling target: aarch64-macos gcc -arch arm64 -shared -std=c11 -O3 -fPIC -I/usr/local/lib/erlang/erts-13.0/include -undefined dynamic_lookup -flat_namespace -undefined suppress /Users/cocoa/Git/fennec_example/c_src/fennec_example.c -o /Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so 20:42:07.678 [debug] Creating precompiled archive: /Users/cocoa/Library/Caches/fennec_precompiled/fennec_example-nif-2.16-aarch64-macos-0.1.0.tar.gz 20:42:07.733 [debug] Restore NIF for current node from: /Users/cocoa/Library/Caches/fennec_precompiled/aarch64-macos.tar.gz After the compilation, you can test the NIF code locally. $ mix test . Finished in 0.01 seconds (0.00s async, 0.01s sync) 1 test, 0 failures Randomized with seed 145253","ref":"precompilation_guide.html#optional-test-the-nif-code-locally","title":"Precompilation guide - (Optional) Test the NIF code locally","type":"extras"},{"doc":"To test precompiling on a local machine, run mix fennec.precompile . $ export FENNEC_CACHE_DIR=&quot;$(pwd)/cache&quot; $ mix fennec.precompile ==&gt; fennec_precompile Compiling 1 file (.ex) 00:25:27.161 [debug] Current compiling target: x86_64-macos ==&gt; fennec_example gcc -arch x86_64 -shared -std=c11 -O3 -fPIC -I/usr/local/lib/erlang/erts-13.0/include -undefined dynamic_lookup -flat_namespace -undefined suppress /Users/cocoa/Git/fennec_example/c_src/fennec_example.c -o /Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so 00:25:27.237 [debug] Creating precompiled archive: /Users/cocoa/Git/fennec_example/cache/fennec_example-nif-2.16-x86_64-macos-0.1.0.tar.gz 00:25:27.286 [debug] Current compiling target: x86_64-linux-gnu zig cc -target x86_64-linux-gnu -shared -std=c11 -O3 -fPIC -I/usr/local/lib/erlang/erts-13.0/include -undefined dynamic_lookup -flat_namespace -undefined suppress /Users/cocoa/Git/fennec_example/c_src/fennec_example.c -o /Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so 00:25:27.346 [debug] Creating precompiled archive: /Users/cocoa/Git/fennec_example/cache/fennec_example-nif-2.16-x86_64-linux-gnu-0.1.0.tar.gz ... Precompiled binaries are stored in the cache subdirectory of the current directory. ls -lah ${FENNEC_CACHE_DIR} total 72 drwxr-xr-x 11 cocoa staff 352B 1 Jul 23:17 . drwxr-xr-x 20 cocoa staff 640B 2 Jul 00:23 .. -rw-r--r-- 1 cocoa staff 1.0K 2 Jul 00:25 fennec_example-nif-2.16-aarch64-linux-gnu-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 994B 2 Jul 00:25 fennec_example-nif-2.16-aarch64-linux-musl-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 1.3K 2 Jul 00:25 fennec_example-nif-2.16-aarch64-macos-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 977B 2 Jul 00:25 fennec_example-nif-2.16-riscv64-linux-musl-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 1.0K 2 Jul 00:25 fennec_example-nif-2.16-x86_64-linux-gnu-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 976B 2 Jul 00:25 fennec_example-nif-2.16-x86_64-linux-musl-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 989B 2 Jul 00:25 fennec_example-nif-2.16-x86_64-macos-0.1.0.tar.gz -rw-r--r-- 1 cocoa staff 4.5K 2 Jul 00:25 fennec_example-nif-2.16-x86_64-windows-gnu-0.1.0.tar.gz drwxr-xr-x 3 cocoa staff 96B 1 Jul 20:49 metadata","ref":"precompilation_guide.html#optional-test-precompiling-locally","title":"Precompilation guide - (Optional) Test precompiling locally","type":"extras"},{"doc":"For this guide the workflow file is located at .github/workflows/fennec_precompile.yml name: precompile on: push: tags: - 'v*' branches: - main jobs: precompile: runs-on: macos-11 env: MIX_ENV: &quot;dev&quot; steps: - uses: actions/checkout@v3 - name: Install Erlang/OTP, Elixir and Zig run: | brew install erlang elixir zig mix local.hex --force mix local.rebar --force - name: Create precompiled library run: | export FENNEC_CACHE_DIR=$(pwd)/cache mkdir -p &quot;${FENNEC_CACHE_DIR}&quot; mix deps.get mix fennec.precompile - uses: softprops/action-gh-release@v1 if: startsWith(github.ref, 'refs/tags/') with: files: | cache/*.tar.gz checksum-*.exs The example here uses macos-11 as the runner because it can generate binaries for most targets.","ref":"precompilation_guide.html#setup-github-actions","title":"Precompilation guide - Setup GitHub Actions","type":"extras"},{"doc":"After CI has finished, you can fetch the precompiled binaries from GitHub. $ mix fennec.fetch --all --print Meanwhile, a checksum file will be generated. In this example, the checksum file will be named as checksum-fennec_example.exs . This checksum file is extremely important in the scenario where you need to release a Hex package using precompiled NIFs. It's MANDATORY to include this file in your Hex package (by updating the files field in the mix.exs ). Otherwise your package won't work . defp package do [ files : [ &quot;lib&quot; , &quot;checksum-*.exs&quot; , &quot;mix.exs&quot; , # ... ] , # ... ] end However, there is no need to track the checksum file in your version control system (git or other).","ref":"precompilation_guide.html#fetch-precompiled-binaries","title":"Precompilation guide - Fetch precompiled binaries","type":"extras"},{"doc":"To recap, the suggested flow is the following: Add :fennec_precompile and relevant fennec_* options to the mix.exs . (Optional) Test if your library compiles locally. mix compile ``` (Optional) Test if your library can precompile to all specified targets locally. mix fennec.precompile ``` Precompile your library on CI. git push origin main --tags ``` Fetch precompiled binaries from GitHub. mix fennec.fetch --all --print ``` Update Hex package to include the checksum file. Release the package to Hex.pm (make sure your release includes the correct files).","ref":"precompilation_guide.html#recommended-flow","title":"Precompilation guide - Recommended flow","type":"extras"},{"doc":"zig seems to not like the -undefined dynamic_lookup and -flat_namespace flag on macOS. Using zig on macOS will cause the build to fail. This is mostly an upstream issue/bug. zig cc -target aarch64-macos -shared -std=c11 -O3 -fPIC -I/usr/local/lib/erlang/erts-13.0/include -L/usr/local/lib/erlang/usr/lib -undefined dynamic_lookup -flat_namespace -undefined suppress /Users/cocoa/Git/fennec_example/c_src/fennec_example.c -o /Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so error(link): undefined reference to symbol &#39;_enif_make_string&#39; error(link): first referenced in &#39;/Users/cocoa/.cache/zig/o/a5316e27e5bd064608113ea75cd212c5/fennec_example.o&#39; error: UndefinedSymbolReference make: *** [/Users/cocoa/Git/fennec_example/_build/dev/lib/fennec_example/priv/nif.so] Error 1","ref":"precompilation_guide.html#notes","title":"Precompilation guide - Notes","type":"extras"}]